#include "internals.h"
#include "locks.h"
#include "hash_table.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdarg.h>
#include <sys/uio.h>

//Internal functions
static DB *_db_alloc(int);
static void _db_free(DB *);
static DB_TBL *_db_tbl_alloc(int);
static DB_TBL *_db_tbl_open(DB *, const char *, tbl_field *, int, int, ...);
static void _db_tbl_free(DB_TBL *);
static DBHASH _db_hash(DB_TBL *, const char *);
static char *_db_encrypt(const char *);
static int	_db_find_and_lock(DB_TBL *, const char *, int);
static void _db_read_dscfile(DB_TBL *db_tbl, char *dsc_filename);
static char *_db_readdat(DB_TBL *);
static off_t _db_readidx(DB_TBL *, off_t);
static off_t _db_readptr(DB_TBL *, off_t);
static void _db_writedat(DB_TBL *, const char *, off_t, int);
static void _db_writeidx(DB_TBL *, const char *, off_t, int, off_t);
static void _db_dodelete(DB_TBL *);
static void _db_writeptr(DB_TBL *, off_t, off_t);
static int _db_findfree(DB_TBL *, int, int);
static void _db_rewind(DB_TBL *);


/* Open/create a database. Arguments are similar to open().
 */
DBHANDLE db_open(const char *pathname, int oflag, ...) {

	DB *db;
	int len, mode;

	len=strlen(pathname);
	if((db = _db_alloc(len)) == NULL) {
		fprintf(stderr, "db_open: _db_alloc error for DB\n.");
		exit(1);
	}

	strcpy(db->name, ".");
	strcat(db->name, pathname);
	strcat(db->name, ".db"); //append index file extension

	if (oflag & O_CREAT) {
		//create mode..so get mode from the variable function arguments
		va_list vargs;
		va_start (vargs, oflag);
		mode = va_arg(vargs, int);
		va_end(vargs);
		
		db->dbfd = open(db->name, oflag | O_RDWR, mode);
		fprintf(stderr, "Creating %s.\n", db->name);
	} else {
		//DB exists. So read the tables from the .db file.
		FILE *file = fopen(db->name, "r");
		char tablename[100];
		DB_TBL *db_tbl;
		int i;
		if (file != NULL) {
			fprintf(stderr, "Reading existing database %s.\n", db->name);
			while (fgets(tablename, sizeof(tablename), file) != NULL) { //read a line
				//simply iterate until we find where the newline is.
				for (i = 0; tablename[i] != NEWLINE; i++);

				tablename[i] = '\0'; //replace newline with null byte.
				db_tbl = _db_tbl_open(db, tablename, NULL, 0, O_RDWR);
				hash_table_insert(db->tables, tablename, db_tbl);
			}
		} else {
			fprintf(stderr, "db_open: Error opening .db file\n");
			exit(1);
		}
		fclose (file);
	
		//open the .db file to get a file descriptor.
		db->dbfd = open(db->name, O_RDWR);
	}

	if (db->dbfd < 0) {
		fprintf(stderr, "db_open: opening %s file failed.\n", db->name);
		_db_free(db);
		return NULL;
	}

	return db;
}

/* Open/create a table. 
 */
DB_TBL* _db_tbl_open(DB* db, const char *tablename, tbl_field *fields, int num_fields, int oflag, ...) {

	DB_TBL *db_tbl;
	int len, mode;
	int i; //iterator

	char descr_string[2*DATLEN_MAX];
	char field_string[DATLEN_MAX];
	char asciiptr[PTR_SZ+1];
	char hash[(NHASH_DEF + 1) * PTR_SZ + 2]; //+2 for null and newline at the end

	struct stat statbuff;

	len=strlen(tablename);
	if((db_tbl = _db_tbl_alloc(len)) == NULL) {
		fprintf(stderr, "_db_tbl_open: _dbtbl_alloc error for DB\n.");
		exit(1);
	}

	db_tbl->nhash = NHASH_DEF;	//Assign hash table size
	db_tbl->hashoff = HASH_OFF; //offset of hash table in index file

	strcpy(db_tbl->name, ".");
	strcat(db_tbl->name, tablename);
	strcat(db_tbl->name, ".idx"); //append index file extension

	if (oflag & O_CREAT) {
		int offset = lseek(db->dbfd, 0, SEEK_END); //seek to end of .db file.
		char table_string[50];
		if (offset < 0) {
			perror("_db_tbl_open: lseek on db file failed");
			exit(1);
		}
		if(writew_lock(db->dbfd, 0, SEEK_SET, 0) < 0) {
			perror("_db_tbl_open: writew_lock failed");
			exit(1);
	    }
		
		strcpy(table_string, tablename);
		strcat(table_string, "\n");
		write(db->dbfd, table_string, strlen(table_string));

		if (un_lock(db->dbfd, 0, SEEK_SET, 0) < 0) {
			fprintf(stderr, "_db_tbl_open: un_lock error on db file.\n");
			exit(1);
		}

		//create mode..so get mode from the variable function arguments
		va_list vargs;
		va_start (vargs, oflag);
		mode = va_arg(vargs, int);
		va_end(vargs);
		
		db_tbl->idxfd = open(db_tbl->name, oflag, mode);
		strcpy(db_tbl->name+(len+1), ".dat");
		db_tbl->datfd = open(db_tbl->name, oflag, mode);
		strcpy(db_tbl->name+(len+1), ".dsc");	
		db_tbl->dscfd = open(db_tbl->name, oflag, mode);
	} else {
		db_tbl->idxfd = open(db_tbl->name, oflag);
		strcpy(db_tbl->name+(len+1), ".dat");
		db_tbl->datfd = open(db_tbl->name, oflag);
		strcpy(db_tbl->name+(len+1), ".dsc");	
		db_tbl->dscfd = open(db_tbl->name, oflag);
		//read existing .dsc file and populate corresponding db_tbl fields.
		_db_read_dscfile(db_tbl, db_tbl->name);
	}

	if (db_tbl->idxfd < 0 || db_tbl->datfd < 0 || db_tbl->dscfd < 0) {
		fprintf(stderr, "_db_tbl_open: opening %s files failed.\n", db_tbl->name);
		_db_tbl_free(db_tbl);
		return NULL;
	}
	
	db->table_count++;

	//if the database was just created, we need to init the .idx file
	if ((oflag & (O_CREAT | O_TRUNC)) == (O_CREAT | O_TRUNC)) {

		//First, initialize the index file.
		if(writew_lock(db_tbl->idxfd, 0, SEEK_SET, 0) < 0) {
			fprintf(stderr, "_db_tbl_open: writew_lock error.\n");
			exit(1);
		}

		if (fstat(db_tbl->idxfd, &statbuff) < 0) {
			fprintf(stderr, "_db_tbl_open: fstat error.\n");
			exit(1);
		}	

		if (statbuff.st_size == 0) {
			sprintf(asciiptr, "%*d", PTR_SZ, 0);
			hash[0] = 0;

			for (i = 0; i < NHASH_DEF + 1; i++) {
				strcat(hash, asciiptr);	
			}
			strcat(hash, "\n");
			if (write(db_tbl->idxfd, hash, strlen(hash)) != strlen(hash)) {
				fprintf(stderr, "_db_tbl_open: index file init write error.\n");
				exit(1);
			}
		}

		if (un_lock(db_tbl->idxfd, 0, SEEK_SET, 0) < 0) {
			fprintf(stderr, "_db_tb_open: un_lock error.\n");
			exit(1);
		}
		
		//Next, initialize the describe file.
		if(writew_lock(db_tbl->dscfd, 0, SEEK_SET, 0) < 0) {
			fprintf(stderr, "_db_tbl_open: writew_lock error.\n");
			exit(1);
		}

		descr_string[0] = 0; //initialize the string for use in strcat.

		//write the number of fields at the start of the dsc record.
		sprintf(field_string, "%d,", num_fields);
		strcat(descr_string, field_string);		

		db_tbl->fields = (tbl_field *)malloc(num_fields * sizeof(tbl_field));
		if(db_tbl->fields == NULL) {
			fprintf(stderr, "_db_tbl_open: malloc error for db_tbl fields.\n");
			exit(1);
		}

		//iterate over the given number of fields.	
		for (i=0; i<num_fields; i++) {
			if(fields[i].name == NULL) {
				fprintf(stderr, "_db_tbl_open: invalid field name.\n");
				exit(1);
			}
			//assign the field struct values to the table's.	
			db_tbl->numfields = num_fields; //assign number of fields in each record in table
			db_tbl->fields[i].type = fields[i].type;
			db_tbl->fields[i].name = (char *)malloc((strlen(fields[i].name)+1) * sizeof(char));
			if(db_tbl->fields[i].name == NULL) {
				fprintf(stderr, "_db_tbl_open: malloc error for db_tbl fieldname.\n");
				exit(1);
			}
			
			strcpy(db_tbl->fields[i].name, fields[i].name);
			db_tbl->fields[i].pos = i;
		
			//format in .dsc file for each field: pos,type,name,pos,type,name,...
			sprintf(field_string, "%d,%d,%s,", (i+1), fields[i].type, fields[i].name);
			strcat(descr_string, field_string);		
		}
		descr_string[strlen(descr_string)-1] = '\n'; //replace the last comma with newline
		descr_string[strlen(descr_string)] = 0; //add null byte to end.

		//copy the descr_string to the describe buffer.
		strcpy(db_tbl->dscbuf, descr_string);
		//also write it to .dsc file
		if (write(db_tbl->dscfd, descr_string, strlen(descr_string)) != strlen(descr_string)) {
			fprintf(stderr, "_db_tbl_open: describe file init write error.\n");
			exit(1);
		}
	
		if(un_lock(db_tbl->dscfd, 0, SEEK_SET, 0) < 0) {
			fprintf(stderr, "_db_tbl_open: un_lock error.\n");
			exit(1);
		}
	}

	_db_rewind(db_tbl);
	return db_tbl;
}

/* Create a new table given a table, information about its fields and the number of fields.
 */
DBHANDLE db_tbl_create(DBHANDLE h, const char *tablename, tbl_field *fields, int num_fields) {

	DB *db = h;
	char *encrypted_tablename = _db_encrypt(tablename);

	DB_TBL *db_tbl = _db_tbl_open(db, encrypted_tablename, fields, num_fields, (O_CREAT | O_TRUNC | O_RDWR), S_IRWXU);
	hash_table_insert(db->tables, encrypted_tablename, db_tbl);

	free(encrypted_tablename);
	return db;
}

/* Allocate and initialize a DB structure
*/
static DB * _db_alloc(int namelen) {
		
	DB *db;
	if((db = malloc(sizeof(DB))) == NULL) {
		fprintf(stderr, "_db_alloc: malloc error.\n");
		exit(1);
	}
	db->dbfd = -1;
	
	db->name = malloc(namelen+5);
	if(db->name == NULL ) {
		fprintf(stderr, "_db_alloc: malloc error for name.\n");
		exit(1);
	}
	db->tables = hash_table_create(0, 0);
	db->table_count = 0;

	return (db);
}

/* Allocate and initialize a DB structure
*/
static DB_TBL * _db_tbl_alloc(int namelen) {
		
	DB_TBL *db_tbl;
	if((db_tbl = calloc(1, sizeof(DB_TBL))) == NULL) {
		fprintf(stderr, "_db_tbl_alloc: calloc error.\n");
		exit(1);
	}
	db_tbl->idxfd = db_tbl->datfd = -1;
	
	db_tbl->name = malloc(namelen+6);
	if(db_tbl->name == NULL ) {
		fprintf(stderr, "_db_tbl_alloc: malloc error for name.\n");
		exit(1);
	}

	db_tbl->idxbuf = malloc(IDXLEN_MAX + 2);
	if (db_tbl->idxbuf  == NULL) {
		fprintf(stderr, "_db_tbl_alloc: malloc error for index buffer.\n");
		exit(1);
	}
	db_tbl->datbuf = malloc(DATLEN_MAX + 2);
	if (db_tbl->datbuf  == NULL) {
		fprintf(stderr, "_db_tbl_alloc: malloc error for data buffer.\n");
		exit(1);
	}
	db_tbl->dscbuf = malloc((DATLEN_MAX*2) + 2);
	if (db_tbl->dscbuf  == NULL) {
		fprintf(stderr, "_db_tbl_alloc: malloc error for describe buffer.\n");
		exit(1);
	}

	return (db_tbl);
}

/* Read an existing .dsc file and populate the table's struct values concerning
 * the fields in the table.
 */
void _db_read_dscfile(DB_TBL *db_tbl, char *dsc_filename) {
	FILE *file = fopen(dsc_filename, "r");
	int i;
	char *token;
	int field_num = 0;
	int num_fields;

	if (file != NULL) {
		while (fgets(db_tbl->dscbuf, (DATLEN_MAX*2 + 2), file) != NULL) { //read line
			//simply iterate until we find where the newline is.
			for (i = 0; db_tbl->dscbuf[i] != NEWLINE; i++);
	
			db_tbl->dscbuf[i] = '\0'; //replace newline with null byte.
		}
		for (i = 0; ;i++, db_tbl->dscbuf = NULL) {
			token = strtok(db_tbl->dscbuf, ",");
			if (token == NULL){
				break;
			}
			if (i == 0) {
				num_fields = atoi(token);
				db_tbl->numfields = num_fields; //assign number of fields in each record in table
				
				db_tbl->fields = (tbl_field *)malloc(num_fields * sizeof(tbl_field));
				if(db_tbl->fields == NULL) {
					fprintf(stderr, "_db_read_dsc: malloc error for db_tbl fields.\n");
					exit(1);
				}
			} else {
				if ((i % 3) == 1) {
					db_tbl->fields[field_num].pos = atoi(token);
				}	
				else if ((i % 3) == 2) {
					db_tbl->fields[field_num].type = atoi(token);
				}	
				else {
					db_tbl->fields[field_num].name = (char *)malloc(strlen(token+1) * sizeof(char));
					if(db_tbl->fields[field_num].name == NULL) {
						fprintf(stderr, "_db_read_dscfile: malloc error for fieldname.\n");
						exit(1);
					}	
					strcpy(db_tbl->fields[field_num].name, token);
					field_num++;
				}
			}
		}
	} else {
		fprintf(stderr, "db_read_dscfile: Error opening .dsc file\n");
		exit(1);
	}
	
	fclose (file);
	return;
}

/*Close database
*/
void db_close(DBHANDLE dbh) {

	_db_free((DB *)dbh);
	return;
} 

/* Free the db struct and other pertaining stuff
 */
static void _db_free(DB *db) {

	char *key;
	DB_TBL *db_tbl;

    hash_table_firstkey(db->tables);
	while(hash_table_nextkey(db->tables, &key, (void **) &db_tbl)) {
		_db_tbl_free(db_tbl);
	} 
    hash_table_delete(db->tables);

	if(db->name != NULL) {
		free(db->name);
	}
	if(db->dbfd >= 0) {
		close(db->dbfd);
	}
	free(db);
	
	return;
}

/* Free the db_tbl struct and other pertaining stuff
 */
static void _db_tbl_free(DB_TBL *db_tbl) {

	if (db_tbl->idxfd >= 0) {
		close(db_tbl->idxfd);
	}
	if (db_tbl->datfd >= 0) {
		close(db_tbl->datfd);
	}
	if (db_tbl->dscfd >= 0) {
		close(db_tbl->dscfd);
	}
	if(db_tbl->idxbuf != NULL) {
		free(db_tbl->idxbuf);
	}
	if(db_tbl->datbuf != NULL) {
		free(db_tbl->datbuf);
	}
	if(db_tbl->dscbuf != NULL) {
		free(db_tbl->dscbuf);
	}
	if(db_tbl->name != NULL) {
		free(db_tbl->name);
	}

	free(db_tbl);	

	return;	
}

/* A simple encryption function.
 */
static char * _db_encrypt(const char *key) {
    char c;
    int i;
    int encrypt_code = 10;
    int key_len = strlen(key);

    char *enc_key = (char *)malloc((key_len+1) * sizeof(char));
    
	for(i=0; (c=*key++)!='\0'; i++){
        enc_key[i] = c+encrypt_code;
	}
   	enc_key[i] = '\0';
 
    return(enc_key);
}


/* Hash value calculation:
 *  	sum the ascii values of the all characters in key
 *		and divide by the number of hash table entries allowed.
 *		The reminder is the hash value for the given key.
 */
static DBHASH _db_hash(DB_TBL *db_tbl, const char *key){
	DBHASH hval = 0;
	char c;
	int i;

	for(i=1; (c=*key++)!=0; i++){
		hval += c*i;
	}
	return(hval % db_tbl->nhash);
}


/* Fetch a record from a table in DB. 
 * Returns pointer to null terminated data.
 */
char * db_fetch(DBHANDLE h, const char * tablename, const char * key) {
	
	DB *db = h;
	DB_TBL *db_tbl;
	
	char *ptr;
	char *encrypted_tablename = _db_encrypt(tablename);

	db_tbl = hash_table_lookup(db->tables, encrypted_tablename);

	if (_db_find_and_lock(db_tbl, key, 0) < 0) {
		ptr = NULL;	
		db_tbl->cnt_fetcherr++;	 //keep count of error
	} else {
		ptr = _db_readdat(db_tbl); //read the data value
		db_tbl->cnt_fetchok++;
	}

	if (un_lock(db_tbl->idxfd, db_tbl->chainoff, SEEK_SET, 1) < 0) {
		fprintf(stderr, "db_fetch: unlock error\n");
		free(encrypted_tablename);
		return NULL;
	}

	free(encrypted_tablename);
	return (ptr);
}

/* Find the specified record. Returns with the hash chain locked.
*/
static int _db_find_and_lock(DB_TBL *db_tbl, const char *key, int writelock) {

	off_t offset, nextoffset;
	
	//calculate the offset in the hash table for this key.
	db_tbl->chainoff = (_db_hash(db_tbl, key) * PTR_SZ) + db_tbl->hashoff;
	db_tbl->ptroff = db_tbl->chainoff;

	//Lock the hash chain here. Lock only the first byte.
	if(writelock) {
		if(writew_lock(db_tbl->idxfd, db_tbl->chainoff, SEEK_SET, 1) < 0) {
			fprintf(stderr, "_db_tbl_find_and_lock: writew_lock error\n");
			return -1;
		}
	} else {
		 if(readw_lock(db_tbl->idxfd, db_tbl->chainoff, SEEK_SET, 1) < 0) {
            fprintf(stderr, "_db_tbl_find_and_lock: readw_lock error\n");
        	return -1;
		}
	}

	//Read the offset value in the first record of the hash chain.
	offset = _db_readptr(db_tbl, db_tbl->ptroff);
	while (offset != 0) {
		nextoffset = _db_readidx(db_tbl, offset);
		if(strcmp(db_tbl->idxbuf, key) == 0) {
			break;
		}
		db_tbl->ptroff = offset; //offset of this unmatched record
		offset = nextoffset; //next one to check
	}	

	if (offset == 0) {
		return -1;
	}
	
	return 0;
}

/* Read a chain ptr field at the given offset. Can read the free list
 * pointer, hash table chain pointer, or index record chain pointer.
 */
static off_t _db_readptr(DB_TBL *db_tbl, off_t offset) {
	
	char asciiptr[PTR_SZ + 1];

	if (lseek(db_tbl->idxfd, offset, SEEK_SET) == -1) {
    	fprintf(stderr, "_db_readptr: lseek error to ptr field\n");
		return -1;	
	}
	if (read(db_tbl->idxfd, asciiptr, PTR_SZ) != PTR_SZ) {
    	fprintf(stderr, "_db_readptr: read error of ptr field\n");
		return -1;	
	}

	asciiptr[PTR_SZ] = 0;
	return(atol(asciiptr));
}

/* Read the index record at offset. Read index record into idxbuf,
 * its corresponding data record offset in the data file to datoff, 
 * and its len to datlen.
 */
static off_t _db_readidx(DB_TBL *db_tbl, off_t offset) {
	ssize_t i;
	char *ptr1, *ptr2;
	char asciiptr[PTR_SZ + 1];
	char asciilen[IDXLEN_SZ + 1];
	struct iovec iov[2];
	
	//position index file and record the offset. 
	//Note: db_nextrec calls with offset = 0 meaning read from current
	//		offset.
	if ((db_tbl->idxoff = lseek(db_tbl->idxfd, offset, offset == 0 ? SEEK_CUR:SEEK_SET))== -1) {
    	fprintf(stderr, "_db_readidx: lseek error\n");
		exit(1);	
	}

	//Read the ascii chain ptr and ascii length at front of the index record.
	iov[0].iov_base = asciiptr;
	iov[0].iov_len = PTR_SZ;
	iov[1].iov_base = asciilen;
	iov[1].iov_len = IDXLEN_SZ;

	if ((i = readv(db_tbl->idxfd, &iov[0], 2)) != PTR_SZ + IDXLEN_SZ) {
    	if (i == 0 && offset == 0) {
			return (-1);
		}
		fprintf(stderr, "_db_readidx: readv error\n");
		exit(1);	
	} 

	asciiptr[PTR_SZ] = 0;
	db_tbl->ptrval = atol(asciiptr); //offset of next key in chain.

	asciilen[IDXLEN_SZ] = 0;
	if((db_tbl->idxlen = atoi(asciilen)) < IDXLEN_MIN || db_tbl->idxlen > IDXLEN_MAX) {
		fprintf(stderr, "_db_readidx: invalid idx record length\n");
		exit(1);
	}
	
	//Read the actual index record.
	if ((i = read(db_tbl->idxfd, db_tbl->idxbuf, db_tbl->idxlen)) != db_tbl->idxlen) {
		fprintf(stderr, "_db_readidx: read error of index record\n");
		exit(1);
	}
	if (db_tbl->idxbuf[db_tbl->idxlen-1] != NEWLINE) {
		fprintf(stderr, "_db_readidx: missing newline\n");
		exit(1);
	}
	db_tbl->idxbuf[db_tbl->idxlen-1] = '\0'; //replace newline with null terminating character
	
	//find separators in the index record.
	if ((ptr1 = strchr(db_tbl->idxbuf, SEP)) == NULL) {
		fprintf(stderr, "_db_readidx: missing first record separator\n");
		exit(1);
	}
	*ptr1++ = 0; //replace SEP with null.

	if ((ptr2 = strchr(ptr1, SEP)) == NULL) {
		fprintf(stderr, "_db_readidx: missing second record separator\n");
		exit(1);
	}
	*ptr2++ = 0;

	if(strchr(ptr2, SEP) != NULL) {
		fprintf(stderr, "_db_readidx: too many record separators\n");
		exit(1);
	}

	//Get the starting offset and length of the corresponding data record
	if((db_tbl->datoff = atol(ptr1)) < 0) {
		fprintf(stderr, "_db_readidx: starting offset < 0\n");
		exit(1);
	}
	if ((db_tbl->datlen = atol(ptr2)) <= 0 || db_tbl->datlen > DATLEN_MAX){
		fprintf(stderr, "_db_readidx: invalid data length\n");
		exit(1);
	}
	
	return (db_tbl->ptrval);
}

/* Read the current data record into the data buffer.
 * Returns a pointer to the null-terminated buffer.
 */
static char * _db_readdat(DB_TBL *db_tbl) {
	if (lseek(db_tbl->datfd, db_tbl->datoff, SEEK_SET) == -1) {
		fprintf(stderr, "_db_readdat: lseek error\n");
		exit(1);
	}
	if (read(db_tbl->datfd, db_tbl->datbuf, db_tbl->datlen) != db_tbl->datlen){
		fprintf(stderr, "_db_readdat: read data error\n");
		exit(1);
	}
	if (db_tbl->datbuf[db_tbl->datlen-1] != NEWLINE) {
		fprintf(stderr, "_db_readdat: missing newline\n");
		exit(1);
	}
	
	db_tbl->datbuf[db_tbl->datlen-1] = 0; //replace newline with null.
	return(db_tbl->datbuf);
}

/* Delete a record pointed by its key value in the table
*/
int db_delete(DBHANDLE h, const char *tablename, const char *key) {
	DB *db = h;
	DB_TBL *db_tbl;
	int rec = 0; //assume record with be found.	

	char *encrypted_tablename = _db_encrypt(tablename);
	db_tbl = hash_table_lookup(db->tables, encrypted_tablename);
	if(db_tbl == NULL){
		fprintf(stderr, "db_delete: hash table lookup for %s (%s) failed\n", encrypted_tablename, tablename);
		exit(1);
	}

	if(_db_find_and_lock(db_tbl, key, 1) == 0) {
		_db_dodelete(db_tbl);
		db_tbl->cnt_delok++;
	} else {
		rec = -1; //not found
		db_tbl->cnt_delerr++;
	}
	
	if(un_lock(db_tbl->idxfd, db_tbl->chainoff, SEEK_SET, 1) < 0) {
		fprintf(stderr, "db_delete: un_lock error\n");
		exit(1);
	}
	
	return (rec);
}


/* Delete the current record pointed in the DB struct.
 * This function is called by db_delete() and db_store() after the record
 * has been located by _db_find_and_lock().
 */
static void _db_dodelete(DB_TBL *db_tbl) {
	int i;
	char *ptr;
	off_t freeptr, saveptr;

	//set data buffer and key to all blanks.
	for(ptr=db_tbl->datbuf, i=0; i < db_tbl->datlen-1; i++) {
		*ptr++ = SPACE;
	}
	*ptr = 0; //null terminate for _db_writedat

	ptr = db_tbl->idxbuf;
	while(*ptr) {
		*ptr++ = SPACE;
	}

	//lock the free list now.
	if(writew_lock(db_tbl->idxfd, FREE_OFF, SEEK_SET, 1) < 0) {
		fprintf(stderr, "_db_dodelete: writew_lock error\n");
		exit(1);
	}

	//write the data record with blanks.
	_db_writedat(db_tbl, db_tbl->datbuf, db_tbl->datoff, SEEK_SET);
	
	//Read the free list pointer. Update its value to the ptr field
	//of this deleted index record (i.e., make deleted record the head 
	//of free list.
	freeptr = _db_readptr(db_tbl, FREE_OFF);

	//Save contents of deleted index record chain ptr before its rewritten.
	saveptr = db_tbl->ptrval;

	//Rewrite the index record to reflect its deleted state. This also rewrites
	//length of index record, data offset and length none of which has changed.
	_db_writeidx(db_tbl, db_tbl->idxbuf, db_tbl->idxoff, SEEK_SET, freeptr);

	//update the free list pointer.
	_db_writeptr(db_tbl, FREE_OFF, db_tbl->idxoff);

	//Rewrite the idx record that pointed to the now deleted record to update
	//its chain pointer field. Recall that _db_find_and_lock sets db_tbl->ptroff
	//to point to this idx record.
	_db_writeptr(db_tbl, db_tbl->ptroff, saveptr);

	//unlock now.
	if(un_lock(db_tbl->idxfd, FREE_OFF, SEEK_SET, 1) < 0) {
		fprintf(stderr, "_db_dodelete: un_lock error\n");
		exit(1);
	}
	
	return;
}


/* Write a data record. Called by _db_dodelete and db_store.
 */
static void _db_writedat(DB_TBL *db_tbl, const char *data, off_t offset, int whence) {
	struct iovec iov[2];
	static char newline = NEWLINE;

	//Lock if we are appending so that lseek and write are atomic. If we are
	//just overwriting, don't need to lock. 
	if(whence == SEEK_END) {//appending..
		if(writew_lock(db_tbl->datfd, 0, SEEK_SET, 0) < 0){
			fprintf(stderr, "_db_writedat: writew_lock error\n");
			exit(1);
		}
	}

	if ((db_tbl->datoff = lseek(db_tbl->datfd, offset, whence)) == -1) {
		fprintf(stderr, "_db_writedat: lseek error\n");
		exit(1);
	}
	db_tbl->datlen = strlen(data) + 1; //add 1 to include newline.

	iov[0].iov_base = (char *) data;
	iov[0].iov_len = db_tbl->datlen - 1;
	iov[1].iov_base = &newline;
	iov[1].iov_len = 1;

	if(writev(db_tbl->datfd, &iov[0], 2) != db_tbl->datlen){
		fprintf(stderr, "_db_writedat: writev error for data record\n");
		exit(1);
	}
	if(whence == SEEK_END) {
		if(un_lock(db_tbl->datfd, 0, SEEK_SET, 0) < 0){
			fprintf(stderr, "_db_writedat: writew_lock error\n");
			exit(1);
		}
	}
	return;
}

/* Write an index record. _db_writedat is called before this to set datoff and
 * datlen fields in the DB_TBL struct.
 */
static void _db_writeidx(DB_TBL *db_tbl, const char *key, off_t offset, int whence, off_t ptrval) {
	struct iovec iov[2];
	char asciiptrlen[PTR_SZ + IDXLEN_SZ + 1];
	int len;
	char *fmt;

	if((db_tbl->ptrval = ptrval) < 0 || ptrval > PTR_MAX) {
			fprintf(stderr, "_db_writeidx: invalid ptr: %ld\n", ptrval);
			exit(1);
	}
	if(sizeof(off_t) == sizeof(long long)){
		fmt = "%s%c%lld%c%d\n";
	}else{
		fmt = "%s%c%ld%c%d\n";
	}
	sprintf(db_tbl->idxbuf, fmt, key, SEP, db_tbl->datoff, SEP, db_tbl->datlen);	
	if((len = strlen(db_tbl->idxbuf)) < IDXLEN_MIN || len > IDXLEN_MAX) {
		fprintf(stderr, "_db_writeidx: invalid idx record length\n");
		exit(1);
	}
	sprintf(asciiptrlen, "%*ld%*d", PTR_SZ, ptrval, IDXLEN_SZ, len);

	//Lock if we are appending so that lseek and write are atomic. If we are
	//just overwriting, don't need to lock. 
	if(whence == SEEK_END) {//appending..
		if(writew_lock(db_tbl->idxfd, ((db_tbl->nhash+1)*PTR_SZ)+1, SEEK_SET, 0) < 0) {
			fprintf(stderr, "_db_writeidx: writew_lock error\n");
			exit(1);
		}
	}

	//Position the index file and record offset.
	if((db_tbl->idxoff = lseek(db_tbl->idxfd, offset, whence)) == -1) {
			fprintf(stderr, "_db_writeidx: lseek error\n");
			exit(1);
	}	
	
	iov[0].iov_base = asciiptrlen;
	iov[0].iov_len = PTR_SZ + IDXLEN_SZ;
	iov[1].iov_base = db_tbl->idxbuf;
	iov[1].iov_len = len;
	
	if(writev(db_tbl->idxfd, &iov[0], 2) != PTR_SZ+IDXLEN_SZ+len) {
		fprintf(stderr, "_db_writedat: writev error for index record\n");
		exit(1);
	}

	if(whence == SEEK_END) { 
		if(un_lock(db_tbl->idxfd, ((db_tbl->nhash+1)*PTR_SZ)+1, SEEK_SET, 0) < 0) {
			fprintf(stderr, "_db_writeidx: un_lock error\n");
			exit(1);
		}
	}

	return; 
}

/* Write a chain ptr field in the free list, the hash table or the
 * index record in the index file.
 */
static void _db_writeptr(DB_TBL *db_tbl, off_t offset, off_t ptrval) {
	char asciiptr[PTR_SZ + 1];
	if(ptrval < 0 || ptrval > PTR_MAX) {
		fprintf(stderr, "_db_writeptr: invalid ptrval\n.");
		exit(1);
	}
	sprintf(asciiptr, "%*ld", PTR_SZ, ptrval);

	if(lseek(db_tbl->idxfd, offset, SEEK_SET) == -1) {
		fprintf(stderr, "_db_writeptr: lseek error\n");
		exit(1);
	}
	if(write(db_tbl->idxfd, asciiptr, PTR_SZ) != PTR_SZ) {
		fprintf(stderr, "_db_writeptr: write error of ptr field.\n");
		exit(1);
	}
	
	return;
}

/* Store a record in DB table. Return 0 if OK, 1 if record exists, -1 on error.
 */
int db_store(DBHANDLE h, const char *tablename, const char *key, const char *data, int flag) {
	DB *db = h;
	DB_TBL *db_tbl;
	int rec;
	int keylen, datlen;
	off_t ptrval;

	char *encrypted_tablename = _db_encrypt(tablename);
	db_tbl = hash_table_lookup(db->tables, encrypted_tablename);

	if(flag!=DB_INSERT && flag!=DB_REPLACE && flag!=DB_STORE) {
		errno = EINVAL;
		return (-1);
	}
	keylen = strlen(key);
	datlen = strlen(data) + 1; //add 1 for newline at end.
	if(datlen < DATLEN_MIN || datlen > DATLEN_MAX) {
		fprintf(stderr, "db_store: invalid data length.\n");
		exit(1);
	}
		
	if(_db_find_and_lock(db_tbl, key, 1) < 0) { //record not found
		if (flag == DB_REPLACE) {
			rec = -1;
			db_tbl->cnt_storerr++;
			errno = ENOENT;	//error for record does not exist.
			goto doreturn;
		}
		//_db_find_and_lock locked the hash chain for us. Now read the chain to
		//the first idx record on the chain.
		ptrval = _db_readptr(db_tbl, db_tbl->chainoff);
		
		if(_db_findfree(db_tbl, keylen, datlen) < 0) {
			//can't find an empty record. Append to end of index and data files.
			_db_writedat(db_tbl, data, 0, SEEK_END);
			_db_writeidx(db_tbl, key, 0, SEEK_END, ptrval);
	
			//The new record goes to the front of the hash chain. db->idxoff is
			//written by _db_writeidx().		
			_db_writeptr(db_tbl, db_tbl->chainoff, db_tbl->idxoff);
			db_tbl->cnt_insert_append++;
		} else {
			//reuse an empty record. _db_findfree() removed it from the free
			//list and set db->datoff and db->idxoff. Again reused record goes
			//to the front of hash chain.
			_db_writedat(db_tbl, data, db_tbl->datoff, SEEK_SET);
			_db_writeidx(db_tbl, key, db_tbl->idxoff, SEEK_SET, ptrval);
			_db_writeptr(db_tbl, db_tbl->chainoff, db_tbl->idxoff);
			db_tbl->cnt_insert_reuse++;
		}
	} else {
		if(flag == DB_INSERT) {
			rec = 1; //record already exists.
			db_tbl->cnt_storerr++;
			goto doreturn;
		}
		
		//We are replacing an exiting record. We know new key equal existing
		//key. But we need to check if the data records are the same size.
		if(datlen != db_tbl->datlen) {
			_db_dodelete(db_tbl); //delete existing record.
			//reread the chain ptr in the hash table.
			ptrval = _db_readptr(db_tbl, db_tbl->chainoff);
			
			//append new index and data records to end of files.
			_db_writedat(db_tbl, data, 0, SEEK_END);	
			_db_writeidx(db_tbl, key, 0, SEEK_END, ptrval);

			//new record goes to front of hash chain.
			_db_writeptr(db_tbl, db_tbl->chainoff, db_tbl->idxoff);
			db_tbl->cnt_replace_append++;
		} else {
			//Same size data. Just replace the record.
			_db_writedat(db_tbl, data, db_tbl->datoff, SEEK_SET);
			db_tbl->cnt_replace_ovwrt++;
		}
	}	

	rec = 0; //everything ok.

  doreturn: //unlock hash chain locked by _db_find_and_lock.
	if(un_lock(db_tbl->idxfd, db_tbl->chainoff, SEEK_SET, 1) < 0) {
		fprintf(stderr, "db_store: un_lock error\n");
		exit(1);
	}
	
	return (rec);
}

/* Try to find a free index and corresponding data record of the given sizes.
 */
static int _db_findfree(DB_TBL *db_tbl, int keylen, int datlen){
	int rec;
	off_t offset, nextoffset, saveoffset;

	//lock free list.
	if(writew_lock(db_tbl->idxfd, FREE_OFF, SEEK_SET, 1) < 0) {
		fprintf(stderr, "_db_findfree: writew_lock error\n");
		exit(1);
	}
	
	//read free list pointer.
	offset = _db_readptr(db_tbl, FREE_OFF);
	while (offset!=0) {
		nextoffset = _db_readidx(db_tbl, offset);
		if(strlen(db_tbl->idxbuf) == keylen && strlen(db_tbl->datbuf) == datlen) {
			break; //found a fitting record.
		}
		saveoffset = offset;
		offset = nextoffset;
	}

	if (offset == 0) {
		rec = -1; //no match found.
	} else {
		//found a free record with matching sizes. saveoffset points to the
		//chain ptr that points to this matching empty record. We now set this
		//chain ptr to db->ptrval to remove the empty record from free list.
		_db_writeptr(db_tbl, saveoffset, db_tbl->ptrval);
		rec = 0;
	}

	//unlock free list.
	if(un_lock(db_tbl->idxfd, FREE_OFF, SEEK_SET, 1) < 0) {
		fprintf(stderr, "_db_findfree: un_lock error\n");
		exit(1);
	}

	return (rec);
}

/* Rewind the index file pointed in db_tbl struct.
 */
void _db_rewind(DB_TBL *db_tbl){
	off_t offset;
	offset = (db_tbl->nhash + 1) * PTR_SZ; //add 1 for free list.
	
	//setting the file offset to the start of the index records.
	if((db_tbl->idxoff = lseek(db_tbl->idxfd, offset+1, SEEK_SET)) == -1){
		fprintf(stderr, "_db_rewind: lseek error to first idx record\n");
		exit(1);
	}
	
	return;
}

/* Rewind the index file for db_nextrec. Must be called before first call of
 * db_nextrec().
 */
void db_rewind(DBHANDLE h, const char *tablename){
	DB *db = h;
	DB_TBL *db_tbl;
	
	char *encrypted_tablename = _db_encrypt(tablename);
	db_tbl = hash_table_lookup(db->tables, encrypted_tablename);

	//call _db_rewind to actually set idxoff to the start of index records.
	_db_rewind(db_tbl);

	return;
}

/* Return the next sequential record in the index file ignoring deleted records.
 * db_rewind() must be called before this is called the first time.
 */
char * db_nextrec(DBHANDLE h, const char *tablename, char *key) {
	DB *db = h;
	char c;
	char *ptr;
	DB_TBL *db_tbl;
	
	char *encrypted_tablename = _db_encrypt(tablename);
	db_tbl = hash_table_lookup(db->tables, encrypted_tablename);

	//Read lock the free list so we don't read a record in the middle that is
	//being deleted.
	if(readw_lock(db_tbl->idxfd, FREE_OFF, SEEK_SET, 1) < 0) {
		fprintf(stderr, "db_writerec: readw_lock error\n");
		exit(1);
	}

	do {
		//read next sequential index record.
		if (_db_readidx(db_tbl, 0) < 0) {
			ptr = NULL;
			goto doreturn;
		}
		
		//check if key is all blank.
		ptr = db_tbl->idxbuf;
 
		//do nothing until we hit null byte or nonblank.
		while ((c=*ptr++)!=0 && c == SPACE)
			;
	} while (c == 0);

	if(key != NULL) {
		strcpy(key, db_tbl->idxbuf); //copy key value.
	}

	ptr = _db_readdat(db_tbl); //return 
	db_tbl->cnt_nextrec++;

  doreturn:
	//unlock the free list.
	if(un_lock(db_tbl->idxfd, FREE_OFF, SEEK_SET, 1) < 0) {
		fprintf(stderr, "db_nextrec: un_lock error\n");
		exit(1);
	}
	
	return (ptr);
}

/* Retrieve all the records in the specified table. And pass the values to the
 * callback function specified by the function pointer.
 */
int db_retrieve(DBHANDLE h, const char *tablename, void *user_data, int (*user_func)(DBHANDLE h, void *user_data, const char *records)) {
	char *key = (char *)malloc(DATLEN_MAX * sizeof(char));
	char *record = (char *)malloc(DATLEN_MAX * sizeof(char));
	char *data = (char *)calloc(1000, DATLEN_MAX);
	if (data == NULL) {
		fprintf(stderr, "db_retrieve: calloc error\n");
		return(-1);
	}	
	strcat (data, "\0");

	db_rewind(h, tablename);
	record = db_nextrec(h, tablename, key);
	while (record != NULL) {	
		strcat(data, key);
		strcat(data, ",");
		strcat(data, record);
		strcat(data, "\n");
		record = db_nextrec(h, tablename, key);
	}
	
	user_func(h, user_data, data);
	return 0;
}

/* Retrieve the tbl_field values in the db_tbl struct. And pass them to the 
 * callback function specified by the function pointer.
 */
int db_getfieldinfo(DBHANDLE h, const char *tablename, void *ud, int (*user_func)(DBHANDLE db, void *ud, int num_fields, tbl_field *)) {
	
	DB *db = h;
	char *encrypted_tablename = _db_encrypt(tablename);
	DB_TBL *db_tbl = hash_table_lookup(db->tables, encrypted_tablename);
	
	if (db_tbl->numfields > 0) {
		if (user_func != NULL) {
			user_func(h, ud, db_tbl->numfields, db_tbl->fields);
		}
	} else {
		fprintf(stderr, "db_getfieldinfo: No fields to return\n");
		return 0;	
	}

	return 1;
}
